// CLRInfo.C

#include "CLRInfo.h"
#include "InputTree.h"
#include <algorithm>
#include "utils.h"
#include <fstream>
#include "FullHist.h"

map<wordtype, map<wordtype, vector<vector<wordtype> > > > CLRInfo::contextMap;

void
CLRInfo::
addCLRPattern (InputTree* tree)
{
  assert (tree->isCLR());
  InputTree* parent = tree->getParent();
  assert (parent);
  if (!tree) cout << tree->toString(); // forces toString to be compiled in
  const wordtype& head = stringFilter(tree->getHead());
  map<wordtype, vector<vector<wordtype> > >& relatedContexts = contextMap[head];
  vector<vector<wordtype> >& verbPrepContexts =
      relatedContexts[stringFilter(parent->getHead())];

  vector<wordtype> newContext;

  for(InputTrees::const_iterator it = parent->getSubTrees().begin();
      it != parent->getSubTrees().end();
      ++it)
  {
    InputTree* child = *it;
    if (tree == child) continue;
    if (child->isCLR() || child->getLabel() == "PRT")
      newContext.push_back(stringFilter(child->getHead()));
  }

  verbPrepContexts.push_back(newContext);
}

bool
CLRInfo::
isCLRContext (const SentRep* sentence, const Hist* tree)
{
  assert (sentence);
  assert (tree);

  static int PRT = -1;
  if (PRT < 0) PRT = Term::get(string("PRT"))->toInt();
  static int VP = -1;
  if (VP < 0) VP = Term::get(string("VP"))->toInt();
  static int S1 = -1;
  if (S1 < 0) S1 = Term::get(string("S1"))->toInt();

  if (tree->getLabel() == PRT) return false;
  if (tree->getLabel() == S1) return false;
  wordtype target = stringFilter(tree->getHeadString());

  const Hist* boundary = tree->getParent();
  assert (boundary);
  while (boundary->getLabel() != VP && boundary->getLabel() != S1)
    boundary = boundary->getParent();
  assert (boundary->getStart() < boundary->getFinish());

  if (!contextMap.count(target)) return false;
  const map<wordtype, vector<vector<wordtype> > >& verbContexts =
    				getByKey(contextMap, target);

  // try the head of each of my ancestors as possible verbs
  for (const Hist* ancestor = tree->getParent();
       ancestor;
       ancestor = ancestor->getParent())
  {
    const wordtype verb = stringFilter(ancestor->getHeadString());

    if (!verbContexts.count(verb)) 
    {
      if (ancestor->getLabel() == VP)
	break;
      else 
        continue;
    }
    const vector<vector<wordtype> >& possibleContexts =
    				  getByKey(verbContexts, verb);

    // foreach possible context
    for (vector<vector<wordtype> >::const_iterator it = possibleContexts.begin();
	 it != possibleContexts.end();
	 ++it)
    {
      const vector<wordtype>& context = *it;

      bool match = true;
      // iterate through the words in that context
      for (vector<wordtype>::const_iterator it = context.begin();
	   it != context.end();
	   ++it)
      {
	const wordtype& element = *it;
	// and see if that word is present in the sentence
	if (!sentence->contains(element, 
				boundary->getStart(),
				boundary->getFinish()))
	{
	  match = false;
	  break;
	}
      }
      if (match) return true;
    }
  }
  return false;
}

void
CLRInfo::
dump (ostream& os)
{
  for (map<wordtype, map<wordtype, vector<vector<wordtype> > > >::const_iterator it = contextMap.begin();
       it != contextMap.end();
       ++it)
  {
    const wordtype& target = (*it).first;
    const map<wordtype, vector<vector<wordtype> > >& verbContexts = (*it).second;
    for (map<wordtype, vector<vector<wordtype> > >::const_iterator it = verbContexts.begin();
	 it != verbContexts.end();
	 ++it)
    {
      const wordtype& verb = (*it).first;
      const vector<vector<wordtype> >& contexts = (*it).second;
      for (vector<vector<wordtype> >::const_iterator it = contexts.begin();
	   it != contexts.end();
	   ++it)
      {
	const vector<wordtype>& context = *it;

	os << "( " << target << "\t" << verb << " ";
	for (vector<wordtype>::const_iterator it = context.begin();
	     it != context.end();
	     ++it)
	{
	  const wordtype& element = *it;
	  os << element << " ";
	}
	os << ")\n";
      }
    }
  }
}

void 
CLRInfo::
init(string data)
{
  init_lemmatiser(data);
  data += "clr.txt";
  ifstream clrStream(data.c_str());
  init(clrStream);
}

void
CLRInfo::
init (istream& is)
{
  string target;
  string verb;
  string nextWord;
  while (is)
  {
    is >> nextWord;
    if (!is) break;
    myassert (nextWord == "(", nextWord);
    is >> target;
    map<wordtype, vector<vector<wordtype> > >& verbContexts = 
      		contextMap[stringFilter(target)];
    is >> verb;
    vector<vector<wordtype> >& contexts = verbContexts[stringFilter(verb)];

    vector<wordtype> newContext;
    while (is)
    {
      is >> nextWord;
      if (nextWord == ")") break;
      newContext.push_back(stringFilter(nextWord));
    }
    contexts.push_back(newContext);
  }
}
