#include "FeatureFunction.h"

#include "Term.h"
#include "Pst.h"
#include "FuncInfo.h"
#include "ccInd.h"

using namespace std;

bool FeatureFunction::initialized = false;
int FeatureFunction::STOP = 0;
int FeatureFunction::S1 = 0;
int FeatureFunction::PP = 0;
int FeatureFunction::DummyWord = 0;
int FeatureFunction::NONE = 0;
string FeatureFunction::datapath;
map<int, const FeatureFunction*> FeatureFunction::index;

void FeatureFunction::init (const string& dpath)
{
  datapath = dpath;

  requireInit<Term>();
  STOP = Term::get(string("STOP"))->toInt();
  S1 = Term::get(string("S1"))->toInt();
  PP = Term::get(string("PP"))->toInt();
  NONE = FuncInfo::NONE;

  requireInit<Pst>();
  DummyWord = Pst::get(string("^^"))->toInt();

  makeFunctions();

  initialized = true;
}

TabularFeatureFunction::TabularFeatureFunction
  (const string& n, int i, const string& file) : FeatureFunction (n, i)
{
  string filename = datapath + file;
  ifstream dataStream(filename.c_str());
  if (!dataStream) error ("Error opening file " + filename);

  string marker;
  dataStream >> marker;
  myassert (marker == "default", marker);

  int dval;
  dataStream >> dval;
  defaultValue_ = dval;

  dataStream >> marker;
  myassert (marker == "parent", marker);

  int parnum;
  dataStream >> parnum;
  parent = FeatureFunction::get(parnum);

  dataStream >> marker;
  if (marker == "bound")
  { 
    dataStream >> lowerBound_;
    dataStream >> upperBound_;
    dataStream >> marker;
  }
  else
  { 
    lowerBound_ = parent->lowerBound();
    upperBound_ = parent->upperBound();
  }

  myassert (marker == "data", marker);

  while (dataStream)
  { 
    int k, v;
    dataStream >> k >> v;
    myassert (table.count(k) == 0, table[k]);
    table[k] = v;
  }
}

const FeatureFunction*
FeatureFunctionDeriver::derive
  (const string& n, int i, const FeatureFunction* parent) const
{ 
  return new DerivedFeatureFunction(n, i, parent, this);
}



class NodeLabel : public CodedFeatureFunction
{
  public:
    NodeLabel(const string& n, int i) : CodedFeatureFunction (n, i) { }
    virtual int apply (const Hist* h) const { return h->getLabel(); }
    virtual int defaultValue() const { return S1; }
    virtual int lowerBound() const { return 0; }
    virtual int upperBound() const { return 78; }
    virtual string valueToString (int i) const
    {
      return Term::fromInt(i)->name();
    }
};

class NodeHeadPOS : public CodedFeatureFunction
{
  public:
    NodeHeadPOS(const string& n, int i): CodedFeatureFunction (n, i) { }
    virtual int apply (const Hist* h) const { return h->getHeadPOS(); }
    virtual int defaultValue() const { return STOP; }
    virtual int lowerBound() const { return 0; }
    virtual int upperBound() const { return 47; }
    virtual string valueToString (int i) const
    {
      return Term::fromInt(i)->name();
    }
};

class NodeHead : public CodedFeatureFunction
{
  public:
    NodeHead(const string& n, int i): CodedFeatureFunction (n, i) { }
    virtual int apply (const Hist* h) const { return h->getHead(); }
    virtual int defaultValue() const { return DummyWord; }
    virtual int lowerBound() const { return -1; }
    virtual int upperBound() const { return 40000; }
    virtual string valueToString (int i) const
    {
      if (i == -1) return "";
      return Pst::fromInt(i);
    }
};

class ParentCCLabel : public CodedFeatureFunction
{
  public:
    ParentCCLabel(const string& n, int i) : CodedFeatureFunction (n, i) { }
    virtual int apply (const Hist* h) const
    {
      const Hist* p = h->getParent();
      if (!p) return defaultValue();
      int parentLabel = p->getLabel();
      if (parentLabel != h->getLabel()) return parentLabel;
      int cc_parentLabel = ltocc(parentLabel, p->isConjunctivePhrase());
      return cc_parentLabel;
    }
    virtual int defaultValue() const { return S1; }
    virtual int lowerBound() const { return 47; }
    virtual int upperBound() const { return 78; }
    virtual string valueToString (int i) const
    {
      return Term::fromInt(i)->name();
    }
};

class FunctionTag : public CodedFeatureFunction
{
  public:
    FunctionTag (const string& n, int i, int f) :
      CodedFeatureFunction (n, i), ftag (f) { }
    virtual int apply (const Hist* h) const
    {
      return h->getFunctionTag(ftag);
    }
    virtual int defaultValue() const { return NONE; }
    virtual int lowerBound() const { return 0; }
    virtual int upperBound() const { return 22; }
    virtual string valueToString (int i) const
    {
      return FuncInfo::fromInt(i)->name();
    }
  private:
    int ftag;
};

class HeadIsAux : public CodedFeatureFunction
{
  public:
    HeadIsAux (const string& n, int i, const string& file) :
      CodedFeatureFunction(n, i)
    {
      string filename = datapath + file;
      ifstream dataStream(filename.c_str());
      if (!dataStream) error ("Error opening file " + filename);

      while (dataStream)
      {
	string pos, word;
	dataStream >> pos >> word;
	if (pos == "") continue;
	assert (word != "");

	int posnum = Term::get(pos)->toInt();
	int wordnum = Pst::get(word)->toInt();
	table[posnum][wordnum] = 1;
      }
    }
    virtual int apply (const Hist* h) const
    {
      int pos = h->getHeadPOS();
      return (table.count(pos) != 0 
	      && getByKey(table,pos).count(h->getHead()) != 0);
    }
    virtual int defaultValue() const { return 0; }
    virtual int lowerBound() const { return 0; }
    virtual int upperBound() const { return 1; }

  private:
    map<int, map<int, int> > table;
};

const Hist* FeatureFunction::calcAltHead(const Hist* hist)
{
  if (hist->getLabel() == PP && hist->getChildren().size() >= 2)
  {
    list<Hist*>::const_iterator it = hist->getChildren().begin();

    for (;
	 it != hist->getChildren().end();
	 ++it)
      if ((*it)->getHead() == hist->getHead()) 
	break;

    if (it == hist->getChildren().end()) return NULL;

    // it is now the preposition that heads the PP.

    ++it;
    if (it == hist->getChildren().end()) return NULL;

    // it is now the object of that preposition.

    return *it;
  }
  else return NULL;
}

class NodeAltHeadPOS : public NodeHeadPOS
{
  public:
    NodeAltHeadPOS (const string& n, int i) : NodeHeadPOS (n, i) { }
    virtual int apply (const Hist* h) const
    {
      const Hist* ah = calcAltHead(h);
      if (ah == NULL) return defaultValue();
      else return ah->getHeadPOS();
    }
};

class NodeAltHead : public NodeHead
{
  public:
    NodeAltHead (const string& n, int i) : NodeHead (n, i) { }
    virtual int apply (const Hist* h) const
    {
      const Hist* ah = calcAltHead(h);
      if (ah == NULL) return defaultValue();
      else return ah->getHead();
    }
};



void FeatureFunction::makeFunctions ()
{
  const FeatureFunctionDeriver* par_d = new ParentDeriver();
  const FeatureFunctionDeriver* gp_d = new GrandparentDeriver();
  const FeatureFunctionDeriver* left_d = new LeftSiblingDeriver();
  const FeatureFunctionDeriver* right_d = new RightSiblingDeriver();

  const FeatureFunction* node_label =
    new NodeLabel ("node label", 0);
  const FeatureFunction* parent_label =
    par_d->derive ("parent label", 1, node_label);
  const FeatureFunction* grandparent_label =
    gp_d->derive ("grandparent label", 9, node_label);
  const FeatureFunction* left_sibling_label =
    left_d->derive ("left sibling label", 7, node_label);
  const FeatureFunction* right_sibling_label =
    right_d->derive ("right sibling label", 8, node_label);
  const FeatureFunction* node_head_pos =
    new NodeHeadPOS ("node head POS", 2);
  const FeatureFunction* parent_head_pos =
    par_d->derive ("parent head POS", 6, node_head_pos);
  const FeatureFunction* grandparent_head_pos =
    gp_d->derive ("grandparent head POS", 10, node_head_pos);
  const FeatureFunction* node_head =
    new NodeHead ("node head", 3);
  const FeatureFunction* parent_head =
    par_d->derive ("parent head", 4, node_head);
  const FeatureFunction* grandparent_head =
    gp_d->derive ("grandparent head", 11, node_head);
  const FeatureFunction* parent_cc_label =
    new ParentCCLabel ("parent cc label", 12);
  const FeatureFunction* syntactic_tag =
    new FunctionTag ("syntactic tag", 25, FuncInfo::GRAMM);
  const FeatureFunction* semantic_tag =
    new FunctionTag ("semantic tag", 26, FuncInfo::FORMFUNC);
  const FeatureFunction* tpc_tag =
    new FunctionTag ("topicalisation tag", 27, FuncInfo::TPC);
  const FeatureFunction* misc_tag =
    new FunctionTag ("miscellaneous tag", 28, FuncInfo::MISC);
  const FeatureFunction* parent_semantic_tag =
    par_d->derive ("parent semantic tag", 32, semantic_tag);
  const FeatureFunction* head_is_aux =
    new HeadIsAux ("head is aux verb", 46, "aux.txt");
  const FeatureFunction* alt_head_pos = 
    new NodeAltHeadPOS ("node alt head POS", 29);
  const FeatureFunction* alt_head = 
    new NodeAltHead ("node alt head", 30);
  const FeatureFunction* left_sibling_head_pos =
    left_d->derive ("left sibling head POS", 47, node_head_pos);
  const FeatureFunction* left_sibling_head =
    left_d->derive ("left sibling head", 48, node_head);
  const FeatureFunction* left_sibling_is_aux =
    left_d->derive ("left sibling is aux verb", 49, head_is_aux);
  const FeatureFunction* right_sibling_head_pos =
    right_d->derive ("right sibling head POS", 50, node_head_pos);
  const FeatureFunction* right_sibling_head =
    right_d->derive ("right sibling head", 51, node_head);
  const FeatureFunction* right_sibling_is_aux =
    right_d->derive ("right sibling is aux verb", 52, head_is_aux);
  const FeatureFunction* parent_is_aux =
    par_d->derive ("parent is aux verb", 53, head_is_aux);
  const FeatureFunction* grandparent_is_aux =
    gp_d->derive ("grandparent is aux verb", 54, head_is_aux);
  const FeatureFunction* two_left_sibling_label =
    left_d->derive ("two left sibling label", 55, left_sibling_label);
  const FeatureFunction* two_right_sibling_label =
   right_d->derive ("two right sibling label", 56, right_sibling_label);
  
}
