#ifndef _FEATUREFUNCTION_H_
#define _FEATUREFUNCTION_H_

#include <string>
#include <map>
#include <list>
#include "Hist.h"
#include "utils.h"

class FeatureFunction 
{
  public:
    int toInt() const { return id; }
    const string& name() const { return name_; }
    virtual int apply (const Hist* h) const = 0;
    virtual int defaultValue () const = 0;
    virtual int lowerBound () const = 0;
    virtual int upperBound () const = 0;
    virtual string valueToString (int i) const
    { 
      return string("") + i;
    }
    string applyStr (const Hist* h) const
    { 
      int val = apply(h);
      myassert (val >= lowerBound(), val);
      myassert (val <= upperBound(), val);
      return valueToString(val);
    }

    static void init (const string& dpath);
    static void makeFunctions ();
    static bool isInitialized() { return initialized; }

    static const FeatureFunction* get (int i)
    {
      if (index.count(i) == 0)
	error (string("FeatureFunction::get: function #") + i 
	    	+ " not found");
      else return index[i];
    }

    typedef map<int, const FeatureFunction*> index_t;
    typedef index_t::const_iterator index_iter_t;
    static index_iter_t begin() { return index.begin(); }
    static index_iter_t end() { return index.end(); }

  protected:
    FeatureFunction (const string& n, int i) : name_(n), id(i)
    {
      if (index.count(i))
	error (string("FeatureFunction: already exists with id ") + i);
      index[i] = this;
    }
    static string datapath;
    static int STOP, S1, PP, DummyWord, NONE;

    static const Hist* calcAltHead(const Hist* hist);

  private:
    string name_;
    int id;
    static bool initialized;

    static index_t index;
};

class TabularFeatureFunction : public FeatureFunction
{ 
  public:
    TabularFeatureFunction
      (const string& n, int i, const string& file);

    virtual int apply (const Hist* h) const
    {
      int parval = parent->apply(h);
      table_iter it = table.find(parval);
      if (it == table.end())
	return defaultValue();
      else
	return (*it).second;
    }

    virtual int defaultValue() const { return defaultValue_; }
    virtual int lowerBound() const { return lowerBound_; }
    virtual int upperBound() const { return upperBound_; }

  private:
    const FeatureFunction* parent;
    int lowerBound_, upperBound_;
    std::map<int,int> table;
    typedef std::map<int,int>::const_iterator table_iter;
    int defaultValue_;
};

class CodedFeatureFunction : public FeatureFunction
{ 
  protected:
    CodedFeatureFunction (const string& n, int i) :
      FeatureFunction(n, i) {  }
};

class DerivedFeatureFunction;

class FeatureFunctionDeriver
{ 
  public:
    const FeatureFunction* derive
     (const string& n, int i, const FeatureFunction* parent) const;
    virtual const Hist* relative (const Hist* h) const = 0;
};

class DerivedFeatureFunction : public FeatureFunction
{ 
  private:
    DerivedFeatureFunction (const string& n, int i,
	const FeatureFunction* p, const FeatureFunctionDeriver* d) :
      FeatureFunction(n, i), parent(p), ffd(d) {  }
    friend class FeatureFunctionDeriver;

  public:
    virtual int apply (const Hist* h) const
    { 
      const Hist* r = ffd->relative(h);
      if (r)
	return parent->apply(r);
      else
	return defaultValue();
    }

    virtual int defaultValue () const { return parent->defaultValue(); }
    virtual int lowerBound () const { return parent->lowerBound(); }
    virtual int upperBound () const { return parent->upperBound(); }

    virtual string valueToString (int i) const
    { 
      return parent->valueToString(i);
    }

  private:
    const FeatureFunction* parent;
    const FeatureFunctionDeriver* ffd;
};

class ParentDeriver : public FeatureFunctionDeriver
{ 
  public:
    virtual const Hist* relative (const Hist* h) const
    { 
      if (h == NULL) return NULL;
      else return h->getParent();
    }
};

class GrandparentDeriver : public FeatureFunctionDeriver
{ 
  public:
    virtual const Hist* relative (const Hist* h) const
    { 
      if (h == NULL || h->getParent() == NULL) return NULL;
      else return h->getParent()->getParent();
    }
};

class LeftSiblingDeriver : public FeatureFunctionDeriver
{ 
  public:
    virtual const Hist* relative (const Hist* h) const
    { 
      const Hist* p = h->getParent();
      if (p == NULL) return NULL;
      
      std::list<Hist*>::const_iterator it = p->getChildren().begin();
      for( ; it != p->getChildren().end() ; it++ )
      {
	const Hist* sibling = *it;
	if (sibling != h) continue;
	if (it == p->getChildren().begin()) return NULL;
	--it;
	sibling = *it;
	assert(sibling);
	return sibling;
      }
      error("LeftSiblingDeriver: parent had no children");
      return NULL;
    }
};

class RightSiblingDeriver : public FeatureFunctionDeriver
{ 
  public:
    virtual const Hist* relative (const Hist* h) const
    { 
      const Hist* p = h->getParent();
      if (p == NULL) return NULL;
      
      std::list<Hist*>::const_iterator it = p->getChildren().begin();
      for( ; it != p->getChildren().end() ; it++ )
      {
	const Hist* sibling = *it;
	if (sibling != h) continue;
	++it;
	if (it == p->getChildren().end()) return NULL;
	sibling = *it;
	assert(sibling);
	return sibling;
      }
      error("RightSiblingDeriver: parent had no children");
      return NULL;
    }
};

#endif
