
#ifndef FEATURETREE_H
#define FEATURETREE_H

#include <cassert>
#include <fstream>
#include <iostream>
#include "Feat.h"
#include "Feature.h"

using namespace std;

class FeatureTree;
class FeatIterator;
class FeatTreeIterator;

#define NULLIND 9999999

class FeatureTree
{
public:
  FeatureTree() 
    : auxNd(NULL), back(NULL), ind_(NULLIND), count(0) {}
  FeatureTree(int i) 
    : ind_(i), auxNd(NULL), back(NULL), count(0) {}
  FeatureTree(int i, FeatureTree* b)
    : ind_(i), auxNd(NULL), back(b), count(0) {}

  void topLevelRead(istream& is);
  int  readOneLevel0(istream& is, int arrayPos);

  /** Selects a sub-featuretree based on the value of a feature.
    * In a feature chain, this will just select the child 
    * corresponding to "when this feature = val".  In a feature
    * tree, there are "auxiliary nodes", so this selects the 
    * auxCntth auxnode. 
    * @param val value of feature
    * @param auxCnt which fork of tree to take */
  FeatureTree* follow(int val, int auxCnt);
  static FeatureTree* roots(int which) { return roots_[which]; }

protected:
  int ind_;
  FeatureTree* back;
  FeatureTree* auxNd;

  int count;

public:
  virtual void init() = 0;
  virtual FeatureTree* getNextFeatureTree(int index, int arrayPos,
      						FeatureTree* parent=NULL) = 0;
  virtual Feat& getNextFeat(int index, int arrayPos) = 0;
  virtual void adjustSize(int numFeats, int numTrees) = 0;
  virtual void performAuxNdVoodoo(int ac) = 0;
  virtual void setFeatureInt(int n) = 0;

  int getInd() const { return ind_; }
  FeatureTree* getParent() { return back; }
  const FeatureTree* getParent() const { return back; }
  int getCount() const { return count; }
  void incCount() { ++count; }

  virtual int getNumChildren() const = 0;
  virtual int getNumFeats() const = 0;

  virtual Feat* getFeat (int id) = 0;
  virtual void incFeat (int id) = 0;
  virtual FeatIterator featsBegin() = 0;
  virtual FeatIterator featsEnd() = 0;

  virtual FeatureTree* getChild (int id) = 0;
  virtual FeatTreeIterator childrenBegin() = 0;
  virtual FeatTreeIterator childrenEnd() = 0;

 friend class FeatTreeIter;
 friend class FeatIter;

 protected:
  static FeatureTree* roots_[NUMCALCS];
 private:
  void read(istream& is, FTypeTree* ftt);
  void othReadFeatureTree(istream& is, FTypeTree* ftt, int cnt);
};

class FeatureIterator 
{ 
  public:
    bool isMapType;
  public:
    virtual Feat* operator*() = 0;
    virtual bool isEqualTo(const FeatureIterator* other) const = 0;
    virtual Feat* operator++() = 0;
};

class FeatIterator
{ 
  private:
    FeatureIterator* it;
  public:
    FeatIterator(FeatureIterator* i) : it(i) {  }
    ~FeatIterator() { delete it; }
    Feat* operator*() { return **it; }
    bool operator==(const FeatIterator& other) const 
      { return it->isEqualTo(other.it); }
    bool operator!=(const FeatIterator& other) const 
      { return !it->isEqualTo(other.it); }
    Feat* operator++() { return ++*it; }
};

class FeatureTreeIterator 
{ 
  public:
    bool isMapType;
  public:
    virtual FeatureTree* operator*() = 0;
    virtual bool isEqualTo(const FeatureTreeIterator* other) const = 0;
    virtual FeatureTree* operator++() = 0;
};

class FeatTreeIterator
{ 
  private:
    FeatureTreeIterator* it;
  public:
    FeatTreeIterator(FeatureTreeIterator* i) : it(i) {  }
    ~FeatTreeIterator() { delete it; }
    FeatureTree* operator*() { return **it; }
    bool operator==(const FeatTreeIterator& other) const 
      { return it->isEqualTo(other.it); }
    bool operator!=(const FeatTreeIterator& other) const 
      { return !it->isEqualTo(other.it); }
    FeatureTree* operator++() { return ++*it; }
};


#endif /* ! FEATURETREE_H */
