#include <fstream>
#include <iostream>
#include <sstream>
#include <cassert>
#include "InputTree.h"
#include "headFinder.h"
#include "FuncInfo.h"
#include "utils.h"

using namespace std;

int            InputTree::pageWidth = 75;  //used for prettyPrinting
int InputTree::printDebug = 0;

InputTree::
~InputTree()
{
  InputTree  *subTree;
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      delete subTree;
    }
  delete treeHist;
}


istream&
operator >>( istream& is, InputTree& parse )
{
  if(parse.word() != "" || parse.term() != ""
     || parse.subTrees().size() > 0)
    error( "Reading into non-empty parse." );
  parse.readParse(is);
  return is;
}


void
InputTree::
readParse(istream& is)
{
  int pos = 0;
  start_ = pos;
  finish_ = pos;

  string temp = readNext(is);
  if(!is) return;
  if(temp != "(") error("Should have seen an open paren here.");
  /* get annotated symbols like NP-OBJ.  term_ = NP ntInfo_ = OBJ */
  temp = readNext(is);
  term_ = "S1";
  if(temp != "(")
    {
      if(temp == "S1" || temp == "TOP")
	{
	  temp = readNext(is);
	}
      else error("readParse: illegal topmost type");
    }
  if(temp != "(") error("readParse: need second open paren here.");
  InputTree* nextTree = newParse(is, pos, this);
  subTrees_.push_back(nextTree);
  list<Hist*> hlist;
  hlist.push_back(nextTree->getTreeHist());
  treeHist = new TreeHist(this, hlist);
  nextTree->getTreeHist()->setParent(treeHist);
  finish_ = pos;
  headTree_ = nextTree->headTree_;
  temp = readNext(is);
  if(temp != ")") error("Should have closed paren here.");
}

InputTree*
InputTree::
newParse(istream& is, int& strt, InputTree* par)
{
  int strt1 = strt;
  string wrd;
  string trm;
  string ntInf;
  bool star = false;
  InputTrees subTrs;

  parseTerm(is, trm, ntInf, star);
#ifdef IGNORE_EXISTING_TAGS
  ntInf = "";
#endif
  for(;;)
    {
      string temp = readNext(is);
      if(temp == "(")
	{
	  InputTree* nextTree = newParse(is, strt, NULL);
	  if(nextTree) subTrs.push_back(nextTree);
	}
      else if(temp == ")") break;
      else
	{
	  //if(trm != "-NONE-")
	    {
	      wrd = temp;
	      strt++;
	    }
	}
    }
  InputTree* ans = new InputTree(strt1, strt, wrd, trm, ntInf, subTrs,
				 par, NULL, false, false, !star);
  InputTreesIter iti = subTrs.begin();
  for(; iti != subTrs.end() ; iti++)
    {
      InputTree* st = *iti;
      st->setParent(ans);
    }
  
  if(wrd == "" && subTrs.size() == 0) return NULL;
  if(wrd != "")
    {
      ans->headTree() = ans;
    }
  else
    {
      int hpos = headPosFromTree(ans);
      ans->headTree() = ithInputTree(hpos, subTrs)->headTree();
    }
  
  return ans;
}

string&
InputTree::
readNext( istream& is ) 
{
  static string tempword[1024];
  static int tempwordnum = 0;

  // if we already have a word, use it, and increment pointer to next word;
  if( tempword[tempwordnum] != "" )
    {
      return tempword[tempwordnum++];
    }
  // else zero out point and stuff in 
  int tempnum;
  for(tempnum = 0 ; tempword[tempnum] != "" ; tempnum++ )
    tempword[tempnum] = "";
  tempwordnum = 0;
  // then go though next input, separating "()[]";
  int    wordnum  = 0 ;
  string  temp;
  is >> temp;
  for( tempnum = 0 ; tempnum < temp.length() ; tempnum++ )
    {
      char tempC = temp[tempnum];
      if(tempC == '(' || tempC == ')' ||
	 tempC == '[' || tempC == ']' )
	{
	  if( tempword[wordnum] != "" )
	    wordnum++;
	  tempword[wordnum++] += tempC;
	}
      else tempword[wordnum] += temp[tempnum];
    }
  return tempword[tempwordnum++];
}

/* if we see NP-OBJ make NP a, and -OBJ b */
void
InputTree::
parseTerm(istream& is, string& a, string& b, bool& star)
{
  string temp = readNext(is);
  if (temp[0] == '*') {
    temp = string(temp, 1);
    star = true;
  }

  if(temp == "(" || temp == ")") error("Saw paren rather than term");
  int len = temp.length();
  size_t pos = temp.find("-");
  /* things like -RCB- will have a - at position 0 */
  if(pos < len && pos > 0)
    {
      string na(temp, 0, pos);
      string nb(temp, pos, len-pos);
      a = na;
      len = pos;
      b = nb;
    }
  else
    {
      a = temp;
      b = "";
    }
  pos = a.find("=");
  if(pos < len && pos > 0)
    {
      string na(temp, 0, pos);
      string nb(temp, pos, len-pos);
      a = na;
      len = pos;
      b += nb;
    }
  pos = a.find("#");
  if(pos < len && pos > 0)
    {
      string na(temp, 0, pos);
      string nb(temp, pos, len-pos);
      a = na;
      b += nb;
    }
  pos = a.find("|");
  if(pos < len && pos > 0)
    {
      string na(temp, 0, pos);
      string nb(temp, pos, len-pos);
      a = na;
      b += nb;
    }
}	   
	   

ostream&
operator <<( ostream& os, const InputTree& parse )
{
  parse.prettyPrint( os, 0, false );
  return os;
}

string
InputTree::
toString ( bool pretty, bool parse ) const
{
  ostringstream outbuf;
  if (pretty && parse)
    prettyPrint (outbuf);
  else if (!pretty && parse)
    printproper (outbuf);
  else //!parse
    printNoParse (outbuf, pretty);
  return string(outbuf.str());
}

int
InputTree::
printNoParse( ostream& os, bool pretty, int start ) const
{
  if( word_.length() != 0 )
  {
    if (pretty && start + word_.length() > pageWidth)
    {
      os << endl;
      start = 0;
    }
    else if (start < 0)
    {
      start = 0;
    }
    else
    {
      os << " ";
      start += 1;
    }
    os << word_;
    start += word_.length();
  }
  else
  {
    bool anyTags = false;
    for (int func = 0; func < 4; ++func)
      if (funcTag[func] != NOFUNC_NUM)
      {
	string tag = FuncInfo::fromInt(funcTag[func])->name();
	if (anyTags)
	{
	  os << "-";
	  start += 1;
	}
	else if (pretty && start + 5 > pageWidth)
	{
	  os << endl << "(";
	  start = 1;
	}
	else if (start < 0)
	{
	  os << "(";
	  start = 1;
	}
	else
	{
	  os << " (";
	  start += 2;
	}
	os << tag;
	start += 3;
	anyTags = true;
      }
    if (!isCorrect())
    {
      os << "*";
      start += 1;
    }

    ConstInputTreesIter  subTreeIter= subTrees_.begin();
    InputTree  *subTree;
    for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      start = subTree->printNoParse( os, pretty, start );
    }
    if (anyTags)
    {
      os << ")";
      start += 1;
    }
  }
  return start;
}

void 
InputTree::
printproper( ostream& os ) const
{
  if( word_.length() != 0 )
    {
      os << "(" << (isCorrect()?"":"*") << term_ << " " << word_ << ")";
    }
  else
    {
      os << "(";
      if (!isCorrect())
        os << "*";
      os <<  term_; // << ntInfo_;
      /*
      if (calculatedComplement && complement_)
	os << "-C";
	*/
      for (int func = 0; func < 4; ++func)
        if (funcTag[func] != NOFUNC_NUM)
	  os << "-" << *FuncInfo::fromInt(funcTag[func]);
      if (hyphnum != -1)
	os << "-" << hyphnum;
      if (eqnum != -1)
	os << "=" << eqnum;
      ConstInputTreesIter  subTreeIter= subTrees_.begin();
      InputTree  *subTree;
      for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
	{
	  subTree = *subTreeIter;
	  os << " ";
	  subTree->printproper( os );
	}
      os << ")";
    }
}

void 
InputTree::
prettyPrint(ostream& os, int start, bool startingLine) const              
{
  if(start >= pageWidth) //if we indent to much, just give up and print it.
  {
    printproper(os);
    return;
  }
  if(startingLine)
    {
      os << "\n";
      int numtabs = start/8;
      int numspace = start%8;
      int i;
      for( i = 0 ; i < numtabs ; i++ ) os << "\t"; //indent;
      for( i = 0 ; i < numspace ; i++ ) os << " "; //indent;
    }
  /* if there is enough space to print the rest of the tree, do so */
  if(spaceNeeded() <= pageWidth-start || word_ != "")
    {
      printproper(os);
    }
  else
    {
      /* we need 2 extra spaces, for "(", " "  */
      int newStart = start + 2 + term_.length();// + ntInfo_.length();
      os << "(";
      if (!isCorrect())
      {
	os << "*";
	++newStart;
      }
      os << term_; // << ntInfo_;
      /*
      if (calculatedComplement && complement_)
      {
	os << "-C";
	newStart += 2;
      }
      */
      for (int func = 0; func < 4; ++func)
        if (funcTag[func] != NOFUNC_NUM)
        {
	  os << "-" << *FuncInfo::fromInt(funcTag[func]);
	  newStart += 4;
        }
      if (hyphnum != -1)
      {
	string hyphstr = string("-") + hyphnum;
	newStart += hyphstr.length();
	os << hyphstr;
      }
      if (eqnum != -1)
      {
	string eqstr = string("=") + eqnum;
	newStart += eqstr.length();
	os << eqstr;
      }
      os << " ";
      //move start to right of space after term_ for next thing to print
      start++; //but for rest just move one space in.
      ConstInputTreesIter  subTreeIter = subTrees_.begin();
      InputTree  *subTree;
      int whichSubTree = 0;
      for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
	{
	  subTree = *subTreeIter;
	  if(whichSubTree++ == 0)
	    {
	      subTree->prettyPrint(os, newStart, false);
	    }
	  else
	    {
	      subTree->prettyPrint(os, start, true);
	    }
	}
      os << ")";
    }
}

/* estimates how much space we need to print the rest of the currently
   print tree */
int
InputTree::
spaceNeeded() const
{
  int needed = 1; // 1 for blank;    
  int wordLen = word_.length();
  if (!isCorrect()) 
    ++needed;	// for *
  needed += wordLen;
  needed += 3; //3 for () and " ";
  needed += term_.length();
  //needed += ntInfo_.length();
  for (int i = 0; i < 4; ++i)
    if (funcTag[i] != NOFUNC_NUM)
      needed += 4;
  if (hyphnum != -1)
  {
    string hyphstr = string("-") + hyphnum;
    needed += hyphstr.length();
  }
  if (eqnum != -1)
  {
    string eqstr = string("=") + eqnum;
    needed += eqstr.length();
  }
  if(word_ != "") return needed;
  ConstInputTreesIter  subTreeIter = subTrees_.begin();
  InputTree  *subTree;
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      needed += subTree->spaceNeeded();
    }
  return needed;
}

void
InputTree::
make(EcSPairs& wordTermList)
{
  if(word_ != "")
      wordTermList.push_back(EcSPair(term_, word_));
  else
    {
      ConstInputTreesIter subTreeIter = subTrees().begin();
      InputTree  *subTree;
      for(; subTreeIter != subTrees().end() ; subTreeIter++)
	{
	  subTree = *subTreeIter;
	  subTree->make(wordTermList);
	}
    }
}

void
InputTree::
make(list<string>& words)
{
  if(word_ != "")
      words.push_back(word_);
  else
    {
      ConstInputTreesIter subTreeIter = subTrees().begin();
      InputTree  *subTree;
      for(; subTreeIter != subTrees().end() ; subTreeIter++)
	{
	  subTree = *subTreeIter;
	  subTree->make(words);
	}
    }
}

InputTree*
ithInputTree(int i, const list<InputTree*> l)
{
  list<InputTree*>::const_iterator li = l.begin();
  for(int j = 0 ; j < i ; j++)
    {
      assert(li != l.end());
      li++;
    }
  return *li;
}

InputTree*
InputTree::
getLeftSibling()
{
  InputTree* par = getParent();
  if (!par) return NULL;
  for (InputTreesIter it = par->getSubTrees().begin();
       it != par->getSubTrees().end();
       ++it)
  {
    InputTree* sibling = *it;
    if (sibling != this) continue;
    if (it == par->getSubTrees().begin()) return NULL;
    --it;
    sibling = *it;
    assert (sibling);
    return sibling;
  }
  error ("getLeftSibling: never found this node!");
}

InputTree*
InputTree::
getRightSibling()
{
  InputTree* par = getParent();
  if (!par) return NULL;
  for (InputTreesIter it = par->getSubTrees().begin();
       it != par->getSubTrees().end();
       ++it)
  {
    InputTree* sibling = *it;
    if (sibling != this) continue;
    if (it == par->getSubTrees().end()) return NULL;
    ++it;
    sibling = *it;
    assert (sibling);
    return sibling;
  }
  error ("getRightSibling: never found this node!");
}
