
#ifndef INPUTTREE_H
#define INPUTTREE_H

#include <list>
#include <string>
#include "TreeHist.h"
#include "FuncInfo.h"

#include "FuncStats.h"
#include "ParseStats.h"
#include "SentRep.h"
#include "utils.h"

using namespace std;

class InputTree;
typedef list<InputTree*> InputTrees;
typedef InputTrees::iterator InputTreesIter;
typedef InputTrees::const_iterator ConstInputTreesIter;
typedef pair<string,string> EcSPair;
typedef list<EcSPair> EcSPairs;
typedef EcSPairs::iterator EcSPairsIter;

class InputTree
{
 public:
  //InputTree(const InputTree& p);
  //InputTree(istream& is);
  InputTree() 
    : start_(0), finish_(0), word_(""), term_(""), parent_(NULL),
  	complement_(false), calculatedComplement(false),
	hyphnum(-1), eqnum(-1), correct(true)
	{ initNode(""); }
  InputTree(int s, int f, const string& w, const string& t,
            const string& n,
	    InputTrees& subT, InputTree* par, InputTree* headTr,
	    bool ccomp=false, bool comp=false, bool corr=true)
    : start_(s), finish_(f), word_(w), term_(t),
      subTrees_(subT), parent_(par), headTree_(headTr), 
      calculatedComplement(ccomp),
      complement_(comp), correct(corr), hyphnum(-1), eqnum(-1)
        { initNode(n); }
      /*
  InputTree(const string w, int i)
    : start_(i), finish_(i+1), word_(w), term_(w),
      parent_(NULL), headTree_(NULL), complement_(false),
      calculatedComplement(false)
	{ memset (funcTag, 0, 4 * sizeof(*funcTag) ); }
	*/
  ~InputTree();

  friend istream& operator >>( istream& is, InputTree& parse );
  friend ostream& operator <<( ostream& os, const InputTree& parse );
  string toString( bool pretty = false, bool parse = true ) const;
  int         start() const { return start_; }
  int         length() const { return (finish() - start()); }
  int         finish() const { return finish_; }
  const string& word() const { return word_; }  
  //string&      word() { return word_; }  
  const string& term() const { return term_; }
  const string& getLabel() const { return term_; }
  //string&      term() { return term_; }
  //const string& ntInfo() const { return ntInfo_; }
  bool isCLR() const 
  { 
    static int CLR = 0;
    if (!CLR) CLR = FuncInfo::get("CLR")->toInt();
    return funcTag[FuncInfo::MISC] == CLR;
  }
  bool isFunc(int type, const string& func) { 
    return funcTag[type] == FuncInfo::get(func)->toInt();
  }
  bool isLOC() const 
  { 
    static int LOC = 0;
    if (!LOC) LOC = FuncInfo::get("LOC")->toInt();
    return funcTag[FuncInfo::FORMFUNC] == LOC;
  }
  void setFuncTag(int i, int val) { funcTag[i] = val; }
  int getFuncTag(int i) { return funcTag[i]; }
  const string head() const { return headTree_->word(); }
  const string getHead() const { return headTree_->word(); }
  const string hTag() const { return headTree_->term(); }
  const string getHeadTag() const { return headTree_->term(); }
  InputTrees& subTrees() { return subTrees_; }
  InputTrees& getSubTrees() { return subTrees_; }
  const InputTrees& getSubTrees() const { return subTrees_; }
  InputTree*& headTree() { return headTree_; }
  void setHeadTree(InputTree* ht) { headTree_ = ht; }
  InputTree*  parent() { return parent_; }
  InputTree*  getParent() { return parent_; }
  void setParent(InputTree* par)
    { parent_ = par; getTreeHist()->setParent(par->getTreeHist()); }
  InputTree* getLeftSibling();
  InputTree* getRightSibling();

  TreeHist* getTreeHist() { return treeHist; }

  void         mapCompat(  InputTree& correct, ParseStats& parseStats,
			   FuncStats& funcStats, SentRep& sr );
  bool         compat( int strt, int fnsh,  SentRep& sr ) ;
  InputTree*   exact( int strt, int fnsh );
  InputTree*   lexact( int strt, int fnsh, const string& trm );
  InputTree*   cexact( int strt, int fnsh, const string& trm,
      					bool cmpl, SentRep& sr);
  InputTree*   funcexact( int strt, int fnsh, const string& trm,
      				int funcVal, SentRep& sr, int func);
  void         mapExact(  InputTree& computed,
			 ParseStats& parseStats, FuncStats& funcStats );
  void         precision(  InputTree& correct, ParseStats& parseStats,
			   FuncStats& funcStats, SentRep& sent,
			   bool restoreFunc=false);
  bool         punctuationClose(int pos1, int pos2,  SentRep& sent);
  InputTree*   lexact2( int strt, int fnsh,
			string trm,  SentRep& sent );
  void         stripFuncTags();
  void         postProcess(int pos);

  bool		isCorrect() const { return correct; }

  static void   init( const string& path );
  static ofstream    funcErrStream;

  bool         isComplement();

  void        make(EcSPairs& str);
  void        make(list<string>& str);
  static int  pageWidth;     
  static int printDebug;

 protected:
  void        readParse(istream& is);
  InputTree*     newParse(istream& is, int& strt, InputTree* par);
  string&  readNext( istream& is );
  void        parseTerm(istream& is, string& a, string& b, bool& star);
  void        printproper( ostream& os ) const;
  void        prettyPrint(ostream& os, int start=0, bool startLine=false) const;
  int         printNoParse ( ostream& os, bool pretty = false, int start=-1 ) const;
  int         spaceNeeded() const;
  int         term_before();
  
  InputTree*  parent_;
  int         start_;
  int         finish_;
  string   word_;
  string   term_;
  bool     complement_;
  bool     calculatedComplement;
  int      funcTag[4];
  int      hyphnum;
  int      eqnum;
  //string   ntInfo_;
  InputTree*  headTree_;
  InputTrees  subTrees_;

  bool     correct;

  TreeHist* treeHist;


  void initNode(const string& ntInfo)
  { 
    list<Hist*> hlist;
    for (InputTrees::iterator it = subTrees_.begin();
	 it != subTrees_.end();
	 ++it)
      hlist.push_back((*it)->getTreeHist());
    treeHist = new TreeHist(this, hlist);

    hyphnum = FuncInfo::getHyphNum(ntInfo);
    eqnum = FuncInfo::getEqNum(ntInfo);
    for (int i = 0; i < 4; ++i)
      funcTag[i] = FuncInfo::getFuncInt(ntInfo, i);
  }
};

InputTree* ithInputTree(int i, const list<InputTree*> l);

#endif /* ! INPUTTREE_H */
