
#include "InputTree.h"
#include <fstream>
#include <iostream>
#include "headFinder.h"
#include "FuncInfo.h"
#include "Term.h"
#include "TreeHist.h"
#include "Feature.h"
#include "ITProb.h"
#include "CLRInfo.h"
#include "parseIntegration.h"
#include "SentRep.h"
#include "utils.h"

//#define TAG_BASELINE
#define TRACK_FUNCERRS

extern SentRep* CurrentSentence;

ofstream       InputTree::funcErrStream;
 
typedef set<int, less<int> > intset;
typedef set<int, less<int> >::iterator intsetiter;

static bool set_has_intersection(intsetiter first1, intsetiter last1,
			intsetiter first2, intsetiter last2);


void
InputTree::
init( const string& path )
{
  funcErrStream.open((path + "funcerr.txt").c_str());
}

void
InputTree::
mapCompat( InputTree& correctTree, ParseStats& parseStats,
	   FuncStats& funcStats, SentRep& sr )
{
  //cerr << "mC " << *this << endl;
  if( subTrees().empty()) return;
  parseStats.totalMap++;
  int    len = finish()-start();
  if( term_ == "S1" )
    {
      parseStats.obviousMap++;
      parseStats.correctMap++;
    }
  else if ( correctTree.compat( start(), finish(), sr ) )
      {
	parseStats.correctMap++;
      }
  InputTree                  *subTree;

  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      subTree->mapCompat( correctTree, parseStats, funcStats, sr );
    }
}
	
bool
InputTree::
compat( int strt, int fnsh, SentRep& sr ) 
{
  //cerr << "compat " << start() << " " << finish()
	 //<< " : " << strt << " " << fnsh << endl;
  int  position = start();
  if( finish() < strt ) return false;
  if( start() > fnsh ) return false;

  bool spc = punctuationClose(start(), strt, sr);
  if( start() > strt && !spc) return false;
  bool fpc = punctuationClose(finish(), fnsh, sr);
  if( finish() < fnsh && !fpc) return false;

  bool      foundStart = false;
  
  InputTreesIter  subTreeIter = subTrees_.begin();
  InputTree    *subTree;    
    
  for( ; ; )
    {
      bool pspc = punctuationClose(position, strt, sr );
      bool pfpc = punctuationClose(position, fnsh, sr );
      if(pspc) foundStart = true;
      if( pfpc )
	if( foundStart ) return true;
      if( position > strt && !foundStart && !pspc ) return false;
      if( position > fnsh && !pfpc ) return false;
      if(subTreeIter == subTrees_.end()) break;
      subTree = *subTreeIter;
      ++subTreeIter;
      if( subTree->compat( strt, fnsh, sr ) )
	return true;
      position = subTree->finish();
    }

  return false;
}
	
InputTree*
InputTree::
exact( int strt, int fnsh )
{
  // first eliminate nodes that are smaller or overlap
  if( start() > strt ) return NULL;
  if( finish() < fnsh ) return NULL;
  /* this next line prevents counting matches against terminals.  Note
     that if we allow traces, then this needs to be made more complex. */
  if(subTrees().empty()) return NULL;  // needed.

  // now see if this node matches
  if( start() == strt && finish() == fnsh ) return this;

  // otherwise recurse downwards and see if a child is an exact match
  InputTree    *subTree;    
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      InputTree* retval;
      if( retval = subTree->exact( strt, fnsh ) )
	return retval;
    }

  // guess not, so return false
  return NULL;
}
	
InputTree*
InputTree::
lexact( int strt, int fnsh, const string& trm )
{
  // first eliminate nodes that are smaller or overlap
  if( start() > strt ) return NULL;
  if( finish() < fnsh ) return NULL;
  /* this next line prevents counting matches against terminals.  Note
     that if we allow traces, then this needs to be made more complex. */
  if(subTrees().empty()) return NULL;  // needed.

  // now see if this node matches
  if( start() == strt && finish() == fnsh && term() == trm ) return this;

  // otherwise recurse downwards and see if a child is an exact match
  InputTree    *subTree;    
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      InputTree* retval;
      if( retval = subTree->lexact( strt, fnsh, trm ) )
	return retval;
    }

  // guess not, so return false
  return NULL;
}

void
InputTree::
mapExact( InputTree& computed, ParseStats& parseStats, FuncStats& funcStats )
{
  if( subTrees().empty() )
    {
      parseStats.ignoredExact++;
      return;
    }
  if(term_ != "" && term_ != "S1")
    {
      /* now this simply records how many applicable constits there are */
      parseStats.totalExact++;
    }
  else parseStats.ignoredExact++;
  
  InputTree                      *subTree;
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      subTree->mapExact( computed, parseStats, funcStats );
    }
}

void
InputTree::
precision( InputTree& correctTree, ParseStats& parseStats,
	   FuncStats& funcStats, SentRep& sent, bool restoreFunc)
{
  static int CLR = -1;
  if (CLR < 0) CLR = FuncInfo::get("CLR")->toInt();

  InputTree* correspondingNode;
  InputTree* lexact1 = NULL;
  if( subTrees().empty() ) // if no subtrees this is a termina symbol;
  {
    parseStats.obviousPrecision++;
    return;
  }
  else if( term_ == "S1" )
  {
    parseStats.obviousPrecision++;

  }
  else 
  {
    parseStats.totalPrecision++;
    if (correspondingNode = correctTree.lexact2( start(), finish(), term_, sent ))
    {
      parseStats.otherWrong++;
      if (lexact1 = correspondingNode->lexact( start(), finish(), term_ ))
      {
	parseStats.exactPrecision++;
	parseStats.correctPrecision++;
      }
      /*
      if (correspondingNode->isComplement() == isComplement())
	parseStats.complementPrecision++;
	*/
      for (int func = 0; func < 4; ++func)
      {
       int correctFuncTag = correspondingNode->getFuncTag(func);
       if (restoreFunc)
       {
        funcTag[func] = correctFuncTag;
       }
       else
       {
	funcStats.witness(correctFuncTag, funcTag[func]);
	if (correctFuncTag == CLR || funcTag[func] == CLR)
	{
	  if (correctFuncTag != funcTag[func])
	  {
	    /*
	    cerr << correctFuncTag << " " << funcTag[func] << " "
	      		<< CLRInfo::isCLRContext(this) << endl;
	    cerr << *getParent() << endl;
	    cerr << "CLR " << CLRInfo::getCLRPattern(correspondingNode)
	      	<< ", " << CLRInfo::getCLRPattern(this) << endl;
	    cerr << *correspondingNode->parent() << endl;
	    cerr << *parent() << endl;
		*/
	  }
	  /*
	  int newCorrTag = (correctFuncTag == CLR) ? 
	    			FuncInfo::lastFuncInfoInt() + 1 + 
				  CLRInfo::getCLRPattern(correspondingNode) : 
				correctFuncTag;
	  int newOutTag = (funcTag[func] == CLR) ? 
	    			FuncInfo::lastFuncInfoInt() + 1 + 
					CLRInfo::getCLRPattern(this) : 
				funcTag[func];
	  funcStats.witness(newCorrTag, newOutTag);
	  */
	}
	if (correctFuncTag == 0 && funcTag[func] == 0)
	{
	  parseStats.funcTagPrecisionA[func]++;
	}
	else if (correctFuncTag == funcTag[func])
	{
	  parseStats.funcTagPrecisionC[func]++;
	}
	else 
	{
	  int rightIndex = correctFuncTag;
	  string rightName(FuncInfo::fromInt(rightIndex)->name());
	  int wrongIndex = funcTag[func];
	  string wrongName(FuncInfo::fromInt(wrongIndex)->name());
	  funcErrStream 
	    << "#Should be " << rightName << ", but marked " << wrongName
	    << ":" << endl << *this << endl;
#ifdef TRACK_FUNCERRS
	  int whichGuessed = GRCALC + func;

	  funcErrStream << "#  my ctxt" << endl;
	  TreeHist* wrongTreeHist = getTreeHist();
	  printFeatTrees(funcErrStream, wrongTreeHist, whichGuessed);
	  float wrongCtxtRightProb =
	    ITProb (rightIndex, wrongTreeHist, whichGuessed);
	  float wrongCtxtWrongProb =
	    ITProb (wrongIndex, wrongTreeHist, whichGuessed);
	  funcErrStream
	    << rightName << "  " << wrongCtxtRightProb << "\t"
	    << wrongName << "  " << wrongCtxtWrongProb << endl;

	  funcErrStream << "#  correct ctxt" << endl;
	  TreeHist* rightTreeHist = correspondingNode->getTreeHist();
	  printFeatTrees(funcErrStream, rightTreeHist, whichGuessed);
	  float rightCtxtRightProb =
	    ITProb (rightIndex, rightTreeHist, whichGuessed);
	  float rightCtxtWrongProb =
	    ITProb (wrongIndex, rightTreeHist, whichGuessed);
	  funcErrStream
	    << rightName << "  " << rightCtxtRightProb << "\t"
	    << wrongName << "  " << rightCtxtWrongProb << endl;

	  if (rightCtxtWrongProb > rightCtxtRightProb) {
	    funcErrStream << "# would have been wrong anyway" << endl;
	  } else {
	    funcErrStream
	      << "# diff (rightright - wrongwrong): " 
	      << rightCtxtRightProb - wrongCtxtWrongProb << endl;
	  }
#endif
	  if (correctFuncTag == 0)
	  {
	    parseStats.funcTagPrecisionE[func]++;
	  }
	  else if (0 == funcTag[func])
	  {
	    parseStats.funcTagPrecisionB[func]++;
	  }
	  else
	  {
	    parseStats.funcTagPrecisionD[func]++;
	  }
	}
       }
      }
    }
    else
      correct = false;
    /* As an optimisation, correctPrecision is counted under 
     * lexact as well. */
    if (!lexact1 && correctTree.exact( start(), finish() ))
    {
      parseStats.correctPrecision++;
    }
  }
#if 0
  else if( correctTree.lexact( start(), finish(), term_ ) )
    {
      lexact1 = true;
      parseStats.correctPrecision++;
      parseStats.otherWrong++;
      parseStats.totalPrecision++;
      parseStats.exactPrecision++;
    }
  else if( correctTree.lexact2( start(), finish(), term_, sent ) )
    {
      parseStats.otherWrong++;
      parseStats.totalPrecision++;
    }
  else parseStats.totalPrecision++;
  if( !lexact1 && term_ != "S1" && correctTree.exact( start(), finish() ) )
    {
      parseStats.correctPrecision++;
    }
  if ( term_ != "S1" && correctTree.cexact( start(), finish(), term_, 
							isComplement(), sent ) )
    {
      parseStats.complementPrecision++;
    }
  if ( term_ != "S1" && correctTree.grammexact( start(), finish(), term_, 
							grammTag, sent ) )
    {
      parseStats.grammTagPrecision++;
    }
#endif
  InputTree                      *subTree;
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      subTree->precision( correctTree, parseStats, funcStats, sent,
	  						restoreFunc);
    }
}

bool
scorePunctuation( const string trmString )
{
  if( trmString == ",") return true;
  if( trmString == ":") return true;
  if( trmString == ";") return true;
  if( trmString == "--") return true;
  if( trmString == "...") return true;
  if( trmString == "-") return true;
  if( trmString == "''") return true;
  if( trmString == "'") return true;
  if( trmString == "`") return true;
  if( trmString == "``") return true;
  if( trmString == ".") return true;
  if( trmString == "?") return true;
  if( trmString == "!") return true;
  return false;
}

/** Returns true iff everything from pos1 to pos2 in the sentence is
  * punctuation */
bool
InputTree::
punctuationClose(int pos1, int pos2, SentRep& sent)
{
  if(pos1 == pos2) return true;
  int i;
  if(pos1 < pos2)
    {
      for(i = pos1 ; i < pos2 ; i++)
	if(!scorePunctuation(sent[i].lexeme())) return false;
      return true;
    }
  else
    {
      for(i = pos2 ; i < pos1 ; i++)
	if(!scorePunctuation(sent[i].lexeme())) return false;
      return true;
    }
  return false;  //Just to make compiler happy;
}
	  
	
// note that lexact2 implies lexact.
InputTree*
InputTree::
lexact2( int strt, int fnsh, string trm, SentRep& sent )
{
  // first eliminate nodes that occupy completely nonoverlapping space
  if( finish() < strt ) return NULL;
  if( start() > fnsh ) return NULL;
  /* this next line prevents counting matches against terminals.  Note
     that if we allow traces, then this needs to be made more complex. */
  if(subTrees().empty()) return NULL;  // needed;

  // now see if this node matches
    // ADVP and PRT are equivalent wrt this metric
    bool termOk = false;
    if(term() == trm) termOk = true;
    else if((term() == "ADVP" || term() == "PRT")
	    && (trm == "ADVP" || trm == "PRT"))
      termOk = true;

    bool startClose = punctuationClose(start(), strt, sent);
    if(!startClose && start() > strt) return NULL;
    bool finishClose = punctuationClose(finish(), fnsh, sent);
    if(!finishClose && finish() < fnsh) return NULL;
    if( startClose && finishClose && termOk ) return this;

  // otherwise recurse downwards and see if a child is an exact match
  InputTree    *subTree;    
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      InputTree* retval;
      if( retval = subTree->lexact2( strt, fnsh, trm, sent ) )
	return retval;
    }

  // guess not, so return false
  return NULL;
}

InputTree*
InputTree::
cexact( int strt, int fnsh, const string& trm,
    					bool cmpl, SentRep& sent )
{
  if( start() > strt ) return NULL;
  if( finish() < fnsh ) return NULL;
  //if( finish() < strt ) return NULL;
  //if( start() > fnsh ) return NULL;
  /* this next line prevents counting matches against terminals.  Note
     that if we allow traces, then this needs to be made more complex. */
  if(subTrees().empty()) return NULL;  // needed.
  bool termOk = false;
  if(term() == trm) termOk = true;
  else if((term() == "ADVP" || term() == "PRT")
	  && (trm == "ADVP" || trm == "PRT"))
    termOk = true;
  bool startClose = punctuationClose(start(), strt, sent);
  if(!startClose && start() > strt) return NULL;
  bool finishClose = punctuationClose(finish(), fnsh, sent);
  if(!finishClose && finish() < fnsh) return NULL;
  if( startClose && finishClose && termOk )//&& isComplement() == cmpl )
    return this;
  InputTree    *subTree;    
    
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      InputTree* retval;
      if( retval = subTree->cexact( strt, fnsh, trm, cmpl, sent ) )
	return retval;
    }
  return NULL;
}

InputTree*
InputTree::
funcexact( int strt, int fnsh, const string& trm, int funcVal, SentRep& sent, int func)
{
  if( start() > strt ) return NULL;
  if( finish() < fnsh ) return NULL;
  /* this next line prevents counting matches against terminals.  Note
     that if we allow traces, then this needs to be made more complex. */
  if(subTrees().empty()) return NULL;  // needed.
  bool termOk = false;
  if(term() == trm) termOk = true;
  else if((term() == "ADVP" || term() == "PRT")
	  && (trm == "ADVP" || trm == "PRT"))
    termOk = true;
  bool startClose = punctuationClose(start(), strt, sent);
  if(!startClose && start() > strt) return NULL;
  bool finishClose = punctuationClose(finish(), fnsh, sent);
  if(!finishClose && finish() < fnsh) return NULL;
  if( startClose && finishClose && termOk && 
      			getFuncTag(func) == funcVal )
    return this;
  InputTree    *subTree;    
    
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      InputTree* retval;
      if( retval = subTree->funcexact( strt, fnsh, trm, funcVal, sent, func ) )
	return retval;
    }
  return NULL;
}

static set<int, less<int> >
buildAdjuncts()
{
  set<int, less<int> > retval;
  retval.insert(FuncInfo::get("ADV")->toInt());
  retval.insert(FuncInfo::get("VOC")->toInt());
  retval.insert(FuncInfo::get("BNF")->toInt());
  retval.insert(FuncInfo::get("DIR")->toInt());
  retval.insert(FuncInfo::get("EXT")->toInt());
  retval.insert(FuncInfo::get("LOC")->toInt());
  retval.insert(FuncInfo::get("MNR")->toInt());
  retval.insert(FuncInfo::get("TMP")->toInt());
  retval.insert(FuncInfo::get("CLR")->toInt());
  retval.insert(FuncInfo::get("PRP")->toInt());
  return retval;
}

int
InputTree::
term_before()
{
  static int stopint = 0;
  if(!stopint)
    {
      string stopnm("STOP");
      stopint = Term::get(stopnm)->toInt();
    }
  InputTree* par = parent();
  if(!par) return stopint;
  InputTreesIter iti = par->subTrees().begin();
  for( ; iti != par->subTrees().end() ; iti++ )
    {
      InputTree* st = *iti;
      if(st != this) continue;
      if(iti == par->subTrees().begin()) return stopint;
      iti--;
      st = *iti;
      const string& trmStr  = st->term();
      const Term* trm = Term::get(trmStr);
      assert(trm);
      return trm->toInt();
    }
  error("Should never get here");
  return -1;
}


bool
InputTree::
isComplement()
{
  if (calculatedComplement)
    return complement_;

  /* Uses the heuristics from Collins97: Complements are either
     1. non terminal that is either
     	1. NP, SBAR[Q], or S, whose parent is S (or SINV or SQ)
	2. NP, SBAR[Q], S, or VP, whose parent is VP
	3. S whose parent is SBAR[Q]
	and
     2. non terminal cannot have func tags ADV, VOC, BNF, DIR,
     	EXT, LOC, MNR, TMP, CLR, or PRP.
     OR
     the first child after the head of a prepositional phrase (PP) */
  static int S = Term::get("S")->toInt();
  static int SQ = Term::get("SQ")->toInt();
  static int SINV = Term::get("SINV")->toInt();
  static int SBAR = Term::get("SBAR")->toInt();
  static int SBARQ = Term::get("SBARQ")->toInt();
  static int NP = Term::get("NP")->toInt();
  static int VP = Term::get("VP")->toInt();
  static int PP = Term::get("PP")->toInt();

  InputTree* par = parent();
  if(!par) return false;
  int parentTerm = Term::get(par->term())->toInt();

  if (parentTerm == PP)
    return (term_before() != Term::get("STOP")->toInt());

  /* If any of the adjunct function tags are on the current tree node,
   * then we must return false.  We perform an intersection between the
   * adjunct function tag set and the set of function tags on this node;
   * if it is not empty, return false. */
  static set<int> adjunctFuncTags = buildAdjuncts();
  set<int> treeFuncTags;
  for (int i = 0; i < 4; ++i)
    if (funcTag[i] != NOFUNC_NUM)
      treeFuncTags.insert(funcTag[i]);
  if (set_has_intersection (adjunctFuncTags.begin(), adjunctFuncTags.end(),
      		    	  treeFuncTags.begin(), treeFuncTags.end()))
    return false;

  int thisTerm = Term::get(term())->toInt();

  if (parentTerm == S || parentTerm == SQ || parentTerm == SINV)
      return (thisTerm == NP || thisTerm == SBAR || thisTerm == SBARQ
	   || thisTerm == S || thisTerm == SQ || thisTerm == SINV);

  else if (parentTerm == VP)
      return (thisTerm == NP || thisTerm == SBAR || thisTerm == SBARQ
	   || thisTerm == S || thisTerm == SQ || thisTerm == SINV
	   || thisTerm == VP);

  else if (parentTerm == SBAR || parentTerm == SBARQ)
      return (thisTerm == S || thisTerm == SQ || thisTerm == SINV);

  else
      return false;
}

void
InputTree::
stripFuncTags ()
{
  if (subTrees().empty()) return;

  for (int i = 0; i < 4; ++i)
    funcTag[i] = NOFUNC_NUM;

  for (InputTreesIter it = subTrees().begin();
       it != subTrees().end();
       ++it)
    (*it)->stripFuncTags();

}

void
InputTree::
postProcess (int pos)
{
  if (subTrees().empty()) return;

  int subpos = 0;
  InputTreesIter  subTreeIter = subTrees_.begin();
  for( ; subTreeIter != subTrees_.end() ; subTreeIter++ )
    {
      (*subTreeIter)->postProcess(subpos++);
    }

  // for now, just determine whether it's a complement or not
  //TreeHist th(this, pos);
  treeHist->setPosition(pos);
  treeHist->setHeadPosition(headPosFromTree(this));  // rationalist heuristic
  /*
  if (ITProb(true, treeHist, FCALC) > 0.5)
    complement_ = true;
    */

  if (getLabel() != "S1")
  {
#ifdef TAG_BASELINE
    funcTag[FuncInfo::GRAMM] = FuncInfo::get("SBJ")->toInt();
    funcTag[FuncInfo::FORMFUNC] = FuncInfo::get("TMP")->toInt();
    funcTag[FuncInfo::TPC] = FuncInfo::get("TPC")->toInt();
    funcTag[FuncInfo::MISC] = FuncInfo::get("TTL")->toInt();
#else
    for (int func = 0; func < 4; ++func)
    {
#ifdef INTEGRATED_FUNCTAG
      if (func == INTEGRATED_FUNCTAG) continue; // handled during parsing
#endif
#ifdef INTEGRATED_FUNCTAG2
      if (func == INTEGRATED_FUNCTAG2) continue; // handled during parsing
#endif
#ifdef CLR_ORACLE
      if (func == FuncInfo::MISC)
      {
	//cerr << "Checking MISC func\n";
	if (CLRInfo::isCLRContext(CurrentSentence, getTreeHist()))
	{
	  //cerr << "Found a CLR context:\n";
	  //cerr << *getParent() << endl;
	  funcTag[func] = FuncInfo::get("CLR")->toInt();
	}
	else
	{
	  funcTag[func] = NOFUNC_NUM;
	}
      }
      else
#endif
      {
	funcTag[func] = NOFUNC_NUM;
	float funcProb = ITProb (funcTag[func], treeHist, GRCALC + func);
	if (printDebug > 10)
	{
	  cerr << "----------------------------------------------" << endl;
	  cerr << *this << endl;
	  if (printDebug > 30)
	    cerr << "no tag : " << funcProb << endl;
	}
	for (intsetiter it = FuncInfo::functionalTags[func].begin();
	     it != FuncInfo::functionalTags[func].end();
	     ++it)
	{
	  float newProb = ITProb (*it, treeHist, GRCALC + func);
	  if (printDebug > 30)
	    cerr << *it << " : " << newProb << endl;
	  if (newProb > funcProb)
	  {
	    funcProb = newProb;
	    funcTag[func] = *it;
	  }
	}
      }
      if (printDebug > 10)
	cerr << "tag : " << funcTag[func] << endl;
    }

#endif
  }
}

static
bool set_has_intersection(intsetiter first1, intsetiter last1,
			intsetiter first2, intsetiter last2) {
    while (first1 != last1 && first2 != last2)
	if (*first1 < *first2)
	    ++first1;
	else if (*first2 < *first1)
	    ++first2;
	else {
	  return true;
	}
    return false;
}
