#include <sys/resource.h>
#include <math.h>
#include <string>
#include <map>
#include "InputTree.h"
#include "ECArgs.h"
#include "headFinder.h"
#include "Term.h"
#include "FuncInfo.h"
#include "CLRInfo.h"
#include "Cluster.h"
#include "FeatureFunction.h"
//#include "FeatureFunctions.h"
//#include "MeChart.h"
#include "Feature.h"
#include "utils.h"

//#define MAXFEATS MAXNUMFS /* defined in Feature.h */
#define MAXFEATS 44 /* because we're not using CLR oracle or parent's rule */

#define USEFUL(X) ((X) < 13 || (X) > 23)


static void featDump (InputTree* tree,
  		      const vector<const FeatureFunction*>& features,
		      const vector<long>& base_indices,
		      const vector<ofstream*>& featFiles);

static bool binaryVectorOutput = true;
static bool skipLeafNodes = false;
static bool skipStarred = false;

inline string usage () {
  return "featDump [options] datadir inputfile+\n"
    	"\t-h\tthis list\n"
	"\t-Fdir\toutput dir (default = \"feat/\")\n"
	"\t-fstem\tSeparate file output; filenames start with stem\n"
	"\t-lN\tskip sentences longer than N words\n"
	"\t-n\tNonterminals only\n"
	"\t-r\tReport only\n"
	"\t-Rfile\tRestrict features dumped to those listed in file\n"
	"\t-s\tSkip bad (starred) constituents\n"
	"\n";
}

int
main(int argc, char *argv[])
{
  ECArgs args( argc, argv, "FflR" );
  if (args.isset('h')) { cout << usage(); exit(0); }

  // prevent core file creation;
  struct rlimit 	core_limits;
  core_limits.rlim_cur = 0;
  core_limits.rlim_max = 0;
  setrlimit( RLIMIT_CORE, &core_limits );

  cerr << "Starting featDump" << endl;

  // Init everything
  if( args.nargs() < 1)
    error( "Need data dir.\n" + usage() );
  string datapath (args.arg(0));
  addSlash (datapath);

  string outputpath ("feat/");
  if (args.isset('F'))
    outputpath = args.value('F');

  string outputStem;
  if (args.isset('f'))
  {
    binaryVectorOutput = false;
    outputStem = args.value('f');
  }
  if (args.isset('s'))
    skipStarred = true;

  map<int,bool> useFeat;
  bool restrictFeats = args.isset('R');
  if (restrictFeats)
  {
    ifstream restrictions(args.value('R').c_str());
    if (!restrictions)
      error (string("featDump: can't open file ") + args.value('R'));
    int featnum;
    while (restrictions)
    {
      restrictions >> featnum;
      useFeat[featnum] = true;
    }
    restrictions.close();
  }
  
  int	maxSentenceLength = 0;
  if (args.isset('l'))
    maxSentenceLength = atoi (args.value('l').c_str());

  if (args.isset('n'))
    skipLeafNodes = true;

  readHeadInfo( datapath );
  FuncInfo::init( datapath );
  Term::init( datapath );
  Cluster::init( datapath );
  Pst pst( datapath );
  init_lemmatiser(datapath);
  //addSubFeatureFunctions();
  FeatureFunction::init(datapath);

  //MeChart::init(datapath);
  //Bchart::setPosStarts();

  vector<string> files;
  if (args.nargs() == 1 && !args.isset('r'))
  {
    // read from stdin?
    error ("Can't yet read from stdin; input file(s) must be provided.");
  }
  else 
  {
    for (int i = 1; i < args.nargs(); ++i)
      files.push_back(args.arg(i));
  }

  // Open output file for each feature, store in vector
  vector<const FeatureFunction*> features;
  vector<ofstream*> featFiles;
  vector<long> base_indices;
  long base_index = 0;
  for (FeatureFunction::index_iter_t it = FeatureFunction::begin();
       it != FeatureFunction::end();
       ++it)
  {
    int id = (*it).first;
    const FeatureFunction* feature = (*it).second;

    if (restrictFeats && !useFeat[id])
      continue;

    features.push_back(feature);
    base_indices.push_back(base_index);
    base_index += feature->upperBound() - feature->lowerBound() + 1;

    if (!binaryVectorOutput)
    {
      string filename (outputpath + outputStem
		  + (id<10? "0": "") + id + ".fv");
      featFiles.push_back(new ofstream(filename.c_str()));
      if (!featFiles.back()->is_open())
	error ("Could not open output file " + filename);

      cerr << "Function " << feature->name() << " going to file "
	<< filename << endl;
    }
  }
  base_indices.push_back(base_index);

  assert (features.size() == base_indices.size() - 1);
  if (!binaryVectorOutput)
    assert (features.size() == featFiles.size());

  if (args.isset('r'))
  {
    for (int i = 0; i < base_indices.size() - 1; ++i)
    {
      cout << strprintf("%2d", features[i]->toInt()) << " ";
      cout << strprintf("%6d", base_indices[i]) << " ";
      cout << strprintf("%-6d", base_indices[i+1] - 1) << " ";
      cout << features[i]->name() << endl;
    }
    exit(0);
  }

  // For each file
    // For each sentence
      // for each constituent
        // for each feature
          // write value of feature to corresponding file
  for (int fnum = 0; fnum < files.size(); ++fnum)
  {
    if (!binaryVectorOutput)
      for (vector<ofstream*>::iterator it = featFiles.begin();
	   it != featFiles.end();
	   ++it)
        if (*it)
          *(*it) << "#From " << files[fnum] << endl;

    ifstream testSStream(files[fnum].c_str());
    if( !testSStream ) error( "No testSStream" );

    for( ; testSStream ; )
    {
      //++sentenceCount;

      InputTree    tree;
      testSStream >> tree;
      if (!testSStream) break;
      if (maxSentenceLength && tree.length() > maxSentenceLength)
      {
	//--sentenceCount;
	continue;
      }

      featDump(&tree, features, base_indices, featFiles);
    }
  }
  return 0;
}

// preorder traversal
static void featDump (InputTree* tree,
  		      const vector<const FeatureFunction*>& features,
		      const vector<long>& base_indices,
		      const vector<ofstream*>& featFiles)
{
  if (skipLeafNodes && tree->getSubTrees().empty()) return;

 if (!skipStarred || tree->isCorrect())
 {
  if (binaryVectorOutput)
  {
    cout << '(';
    for (int i = 0; i < features.size(); ++i)
      cout
	<< (i?" ":"")
	<< features[i]->apply(tree->getTreeHist()) 
	   + base_indices[i] - features[i]->lowerBound();
    cout << ")\n";
  }
  else for (int i = 0; i < features.size(); ++i)
  {
    *featFiles[i] << features[i]->applyStr(tree->getTreeHist()) << endl;
  }
 }

  for (InputTrees::iterator it = tree->getSubTrees().begin();
       it != tree->getSubTrees().end();
       ++it)
    featDump (*it, features, base_indices, featFiles);
}
