#include "featRead.h"

#include <vector>
#include <map>
#include <algorithm>
#include <functional>
#include "ECArgs.h"
#include "BracketString.h"
#include "SVectorSparse.h"
#include "FuncInfo.h"
#include "Term.h"
#include "Pst.h"
#include "FeatureFunction.h"
#include "utils.h"

using namespace std;

static string usage() {
  return "featRead datadir inputfile(s)\n"
    	"\t-h\tthis list\n"
    	"\t-Rfile\tRestrict features to those listed in file\n"
	"\n";
}

struct print_feat_name : public unary_function<int,void>
{
  print_feat_name(
  	const vector<const FeatureFunction*>& f,
  	const vector<long>& b) :
    features(f), base_indices(b), feat(0) { }
  const vector<const FeatureFunction*>& features;
  const vector<long>& base_indices;
  int feat;
  void operator()(pair<int,int> idpair) 
  {
    int id = idpair.first;
    if (id < base_indices[feat]) feat = 0;
    while (feat < base_indices.size() && base_indices[feat] <= id)
      ++feat;
    --feat;
    myassert (feat >= 0, feat);
    myassert (feat < base_indices.size(), feat);
    int featdiff = id - base_indices[feat];
    int featval = features[feat]->lowerBound() + featdiff;

    cout << "  "
      << features[feat]->name() << '='
      << features[feat]->valueToString(featval) << endl;
  }
};

int main (int argc, char *argv[])
{
  ECArgs args(argc, argv, "R");
  if (args.isset('h')) { cout << usage(); exit(0); }

  {
    char *argNames[] = {"data directory", "input file(s)"};
    checkMinArgs (args.nargs(), 2, argNames);
  }

  map<int,bool> useFeat;
  bool restrictFeats = args.isset('R');
  if (restrictFeats)
  {
    ifstream restrictions(args.value('R').c_str());
    if (!restrictions)
      error (string("featDump: can't open file ") + args.value('R'));
    int featnum;
    while (restrictions)
    {
      restrictions >> featnum;
      useFeat[featnum] = true;
    }
    restrictions.close();
  }

  string datapath = args.arg(0);
  addSlash (datapath);

  FuncInfo::init( datapath );
  Term::init( datapath );
  Pst pst( datapath );
  FeatureFunction::init( datapath );

  vector<const FeatureFunction*> features;
  vector<long> base_indices;
  long base_index = 0;
  for (FeatureFunction::index_iter_t it = FeatureFunction::begin();
       it != FeatureFunction::end();
       ++it)
  {
    int id = (*it).first;
    const FeatureFunction* feature = (*it).second;

    if (restrictFeats && !useFeat[id])
      continue;

    features.push_back(feature);
    base_indices.push_back(base_index);
    base_index += feature->upperBound() - feature->lowerBound() + 1;
  }

  print_feat_name printer(features, base_indices);


  for (int i = 1; i < args.nargs(); ++i)
  {
    ifstream input(args.arg(i).c_str());
    if (!input) error ("Bad input filename " + args.arg(i));
    cerr << endl << "Reading from " << args.arg(i) << "..." << endl;

    while (input)
    {
      BracketString bstr;
      input >> bstr;
      SVectorSparse datum;
      datum.readString(bstr);

      cout << "---" << endl;
      for_each (datum.begin(), datum.end(), printer);
    }
  }
}
