/*
 * Copyright 1997, Brown University, Providence, RI.
 * 
 *                         All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose other than its incorporation into a
 * commercial product is hereby granted without fee, provided that the
 * above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Brown University not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 * 
 * BROWN UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL BROWN UNIVERSITY BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef UTILS_H 
#define UTILS_H

#include <string>
#include <cstdlib>
#include <cassert>
#include <cstdio>
#include <fstream>
#include <list>
#include <vector>
#include <map>
#include <sstream>
#include <iostream>

extern int errno;

//using namespace std;

inline void error(const char *s)
  { std::cerr << "error: " << s << std::endl; abort(); exit(1); }
inline void error(const std::string s) { error (s.c_str()); }
inline void warn(const char *s)
  { std::cerr << "warning: " << s << std::endl; }
inline void warn(const std::string s) { warn (s.c_str()); }
double ran();
double mylog2(double n);
char* toLower(const char* str, char* temp);
std::string toLower(std::string str);
std::string lemmatise(const std::string& str);
void init_lemmatiser(std::string datadir);

enum DGStatus  { OK, NOT_OK };

void ignoreComment(std::ifstream& input);

template <class T>
T
ith(int i, const std::list<T>& l)
{
  typename std::list<T>::const_iterator li = l.begin();
  for(int j = 0 ; j < i ; j++)
    {
      assert(li != l.end());
      li++;
    }
  return *li;
}


#ifdef NDEBUG
#define myassert(X,Y) /* do nothing */
#else
#define myassert(X,Y) if(!(X)){ std::cerr << "data is " << (Y) << std::endl; assert(X); }
#endif

/*
template <class Data>
inline std::string operator+(const std::string &str, const Data &data) {
  std::ostringstream os(str);
  os << data;
  return os.str();
}
*/
inline std::string operator+(const std::string &str, const int& num) {
  std::ostringstream os;
  os << str << num << '\0';
  return std::string(os.str());
}

template<class T>
inline static bool
vectorContains (const std::vector<T>& vec, const T& item)
{
  return vec.end() != find(vec.begin(), vec.end(), item);
}

// Does what [] should do, but doesn't.  Can't overload [] because C is stupid.
template<class A, class B>
inline static const B&
getByKey (const std::map<A, B>& themap, const A& thekey)
{
  typename std::map<A,B>::const_iterator it = themap.find(thekey);
  //if (it == themap.end()) error ("Key not found in map");
  assert (it != themap.end());
  return (*it).second;
}

// Does what [] should do, but doesn't.  Can't overload [] because C is stupid.
template<class A>
inline A&
vecRef (std::vector<A>& thevec, int i)
{ 
  if (i + 1 > thevec.size())
    thevec.resize(i + 1);
  return thevec[i];
}

inline long
stringToLong(const std::string& str)
{ 
  long retval = strtol (str.c_str(), NULL, 10);
  if (retval == 0  &&  errno)
    throw errno;
  else return retval;
}

static long intpowaux (int m, int n, int mult = 1);
inline long intpow(int m, int n, int mult = 1)
{ 
  assert (n >= 0);
  switch (n)
  { 
    case 0: return mult;
    case 1: return m * mult;
    case 2: return m*m * mult;
    default: return intpowaux(m,n,mult);
  }
}

static long intpowaux (int m, int n, int mult)
{ 
  if (n % 2 == 0) //even
    return intpow (m * m, n / 2, mult);
  else //odd
    return intpow (m * m, n / 2, mult * m);
}


template<class A>
inline void requireInit()
{ 
  if (!A::isInitialized())
  { 
    error ("Some class was not initialized (use gdb to diagnose)");
  }
}

inline void
checkMinArgs (int actual, int should, char* names[])
{
  if (actual < should)
    error (std::string("Too few arguments: supply ") + names[actual] + ".");
}

inline void
addSlash (std::string& dirname)
{ 
  if (dirname[dirname.length()-1] != '/')
    dirname += "/";
}

// update this eventually for unlimited args
template<class T>
inline std::string
strprintf (const char* fmt, T arg1)
{
  char buf[100];
  snprintf (buf, 100, fmt, arg1);
  return std::string(buf);
}


#endif /* ! UTILS_H */
